from __future__ import annotations

import asyncio
import logging
import random
import string
import warnings
from typing import Iterable, Optional

import carehare
from channels.exceptions import ChannelFull, StopConsumer
from channels.layers import BaseChannelLayer

from .multiqueue import MultiQueue
from .reader import consume_into_multi_queue_until_connection_close
from .util import (
    ChannelRecipient,
    GroupRecipient,
    gather_without_leaking,
    serialize_message,
)

logger = logging.getLogger(__name__)

ReconnectDelay = 1.0  # seconds


async def _setup_connection(
    *,
    connection: carehare.Connection,
    queue_name: str,
    remote_capacity: Optional[int],
    expiry: Optional[int],
    groups_exchange: str,
) -> None:
    # Declare "groups" exchange. It may persist; spurious declarations
    # (such as on reconnect) are harmless.
    await connection.exchange_declare(groups_exchange, exchange_type="direct")

    # If we're reconnecting to RabbitMQ Node B after Node A disconnected us, our
    # previously-declared exclusive queue may still exist but _won't_ raise a
    # "resource-locked" error.
    #
    # [adamhooper, 2021-06-07] I'm not 100% sure this can actually happen. We
    # saw a problem that _looks_ like this on production, twice, when we
    # upgraded our node pools. The `queue_declare()` succeeded, but subsequent
    # `group_add()` operations failed because the queue did not exist. My theory
    # is: halfway along the way to deleting the exclusive queue, queue_declare()
    # succeeds, because maybe there's a race when deleting an exclusive queue in
    # which it appears non-exclusive to other nodes before it's deleted?
    #
    # ... this guess doesn't _literally_ violate the spec: "The client MAY NOT
    # attempt to use a queue that was declared as exclusive by another
    # still-open connection." Doesn't say anything about queues declared as
    # exclusive by another _closed_ connection on another node.
    #
    # If our woes persist, or if we go 1yr without seeing this message in
    # our logs, delete this block because we'll know it serves no purpose.
    # If, 1yr from now, we've solved our problem, let's build a reproduction
    # and report this upstream.
    try:
        await connection.queue_declare(queue_name, passive=True)
        # If we got here, the queue exists. The `queue_declare()` _below_ ought
        # to raise a 405 error if the queue still exists by the time we call it;
        # but we're guessing it might not, and there's no harm in raising the
        # same error ourselves.
        #
        # This exception will lead to a log message and a reconnect.
        #
        # TODO test using a whole RabbitMQ cluster, to expose this bug and
        # test this solution.
        raise carehare.ChannelClosedByServer(
            405, "channels_rabbitmq refusing to redeclare queue %s" % queue_name
        )
    except carehare.ChannelClosedByServer as err:
        if err.reply_code != 404:
            raise

    arguments = {}
    if remote_capacity is not None:
        arguments["x-max-length"] = remote_capacity
        arguments["x-overflow"] = "reject-publish"
    if expiry is not None:
        arguments["x-message-ttl"] = int(expiry * 1000)
    # This is an exclusive queue, because we want it to disappear when we
    # shut down. (Otherwise we'd leak queues every deploy.) And since an
    # exclusive queue disappears when we aren't connected, there's no way
    # to preserve every message across connects. That's okay -- disconnect
    # means one or two lost messages and nothing more.
    await connection.queue_declare(queue_name, exclusive=True, arguments=arguments)


async def _setup_subscriptions(
    *,
    connection: carehare.Connection,
    queue_name: str,
    groups_exchange: str,
    groups: Iterable[str],
) -> None:
    await gather_without_leaking(
        connection.queue_bind(
            queue_name=queue_name, exchange_name=groups_exchange, routing_key=group
        )
        for group in groups
    )


def _assert_channel_is_reply_channel(channel_name: str) -> None:
    if "!" not in channel_name:
        raise RuntimeError(
            """
***
You are trying to use "Normal Channels", as per
https://channels.readthedocs.io/en/1.x/concepts.html#channel-types.

channels_rabbitmq does not support Normal Channels. See
https://github.com/CJWorkbench/channels_rabbitmq/pull/11#issuecomment-499185070
for an explanation. You can build a pull request if you really want
this; but we heartily recommend you implement your work queue using
carehare directly. Your service will be simpler and it will scale better.
***"""
        )


def _random_letters(n: int) -> str:
    return "".join(random.choice(string.ascii_letters) for i in range(n))


class RabbitmqChannelLayer(BaseChannelLayer):
    """RabbitMQ channel layer.

    It routes all messages into a remote RabbitMQ server.

    Queue names look like "channels_asldkjfg"; Channel names look like
    "channels_asldkjfg!asdaSDGdlkgj". RabbitMQ only sees the queue name; the
    channel names are embedded in `message["__asgi_channel__"]`. Each
    Connection is responsible for a single queue and multiple channels on that
    queue. We call that queue a "channel_key".

    There are two "capacity" parameters: `remote_capacity` determines how many
    messages RabbitMQ will queue before a send raises `ChannelFull`.
    `local_capacity` determines the maximum number of messages to hold in
    memory in our ever-running pull-messages-from-RabbitMQ loop. If consumers
    are too slow then `local_capacity` messages will be buffered in memory; at
    that point Connection won't pull any more from RabbitMQ; so senders can
    send another `remote_capacity` messages to RabbitMQ and then they'll start
    raising `ChannelFull`.

    There is also an "expiry" parameter: this determines the minimum number of
    seconds a message must remain in RabbitMQ before being culled. A similar
    `local_expiry` determines the minimum amount of time a message must remain
    in the channels_rabbitmq buffers before it is acked (and potentially not
    delivered to all consumers).

    This layer does not implement "flush". To flush all state, simply
    disconnect all clients.
    """

    extensions = ["groups"]

    def __init__(
        self,
        host="amqp://guest:guest@127.0.0.1/asgi",
        local_capacity=100,
        remote_capacity=100,
        expiry=60,
        local_expiry: Optional[float] = None,
        group_expiry=None,
        ssl_context=None,
        groups_exchange="groups",
    ):
        self.host = host
        self.connect_timeout = 10.0  # seconds
        self.local_capacity = local_capacity
        self.remote_capacity = remote_capacity
        self.expiry = expiry
        self.local_expiry = local_expiry
        self.ssl_context = ssl_context
        self.groups_exchange = groups_exchange

        try:
            asyncio.get_running_loop()
        except RuntimeError:
            raise RuntimeError(
                """Refusing to initialize channel layer without a running event loop.

If you're writing a django-channels Worker, channels_rabbitmq does not support
this. Read the way to write workers in the carehare documentation.

If you're calling `async_to_sync()`, the call must be within code run by
`sync_to_async()`. Django Channels guarantees this for Django views. Elsewhere,
beware: `asgiref` *does* let you to call `async_to_sync()` without
`sync_to_async()`, but `channels_rabbitmq` *doesn't* -- hence this error."""
            )

        if group_expiry is not None:
            warnings.warn(
                (
                    "channels_rabbitmq does not support group_expiry. Please do not configure it. "
                    "For rationale, see "
                    "https://github.com/CJWorkbench/channels_rabbitmq/issues/18#issuecomment-547052373"
                ),
                category=DeprecationWarning,
            )

        # Choose queue name here: that way we can declare it on RabbitMQ with
        # exclusive=True and have it survive reconnections.
        self._queue_name = "channels_{rand}".format(rand=_random_letters(12))

        self._multi_queue = MultiQueue(capacity=local_capacity)
        self._carehare_connection: asyncio.Future[
            carehare.Connection
        ] = asyncio.Future()
        self._want_close: bool = False

    @property
    def carehare_connection(self) -> asyncio.Future[carehare.Connection]:
        if self._want_close:
            raise StopConsumer
        if self._carehare_connection.get_loop() != asyncio.get_running_loop():
            raise RuntimeError(
                "The caller tried using channels_rabbitmq on a different event loop than the one it was initialized with."
            )
        if not self._want_close and not hasattr(self, "_reconnect_forever_task"):
            self._reconnect_forever_task = asyncio.create_task(
                self._reconnect_forever()
            )
            if self.local_expiry is not None:
                self._expire_task = asyncio.create_task(
                    self._multi_queue.expire_locally_until_closed(
                        local_expiry=self.local_expiry
                    )
                )
        return self._carehare_connection

    async def _reconnect_forever(self):
        """Maintain self._carehare_connection.

        Cancel this task on event-loop shutdown (and at no other time). This
        will make `self.carehare_connection` raise `asyncio.CancelledError`.
        """
        EXPECTED_EXCEPTIONS = (
            carehare.ConnectionClosedByServer,
            carehare.ConnectionClosedByHeartbeatMonitor,
            OSError,
        )

        while not self._want_close:  # Reconnect (after disconnect) repeatedly
            while not self._want_close:  # Retry initial connect repeatedly
                try:
                    connection = carehare.Connection(
                        url=self.host,
                        connect_timeout=self.connect_timeout,
                        ssl=self.ssl_context,
                    )
                    await connection.connect()
                except (asyncio.TimeoutError, *EXPECTED_EXCEPTIONS) as err:
                    logger.exception(
                        "Failure connecting to RabbitMQ: %s. Retrying", str(err)
                    )
                    await asyncio.sleep(ReconnectDelay)
                    connection = None  # in case _want_close=True
                    continue  # retry

                try:
                    await _setup_connection(
                        connection=connection,
                        queue_name=self._queue_name,
                        remote_capacity=self.remote_capacity,
                        expiry=self.expiry,
                        groups_exchange=self.groups_exchange,
                    )
                    await _setup_subscriptions(
                        connection=connection,
                        queue_name=self._queue_name,
                        groups_exchange=self.groups_exchange,
                        groups=self._multi_queue._local_groups.keys(),
                    )
                except (carehare.ChannelClosedByServer, *EXPECTED_EXCEPTIONS) as err:
                    logger.exception(
                        "Failure declaring RabbitMQ objects: %s. Will retry.",
                        str(err),
                    )
                    await asyncio.sleep(ReconnectDelay)
                    connection = None  # in case _want_close=True
                    continue  # retry

                self._carehare_connection.set_result(connection)
                logger.info("Connected to RabbitMQ")
                break

            if self._want_close:
                break

            try:
                await consume_into_multi_queue_until_connection_close(
                    connection=connection,
                    channel=self._queue_name,
                    multi_queue=self._multi_queue,
                    prefetch_count=self.local_capacity,
                )
                await connection.closed
            except EXPECTED_EXCEPTIONS as err:
                logger.exception(
                    "Disconnected from RabbitMQ: %s. Will reconnect.", str(err)
                )

            self._carehare_connection = asyncio.Future()
            try:
                await connection.close()
            except EXPECTED_EXCEPTIONS:
                pass
            connection = None

        if connection is not None:
            try:
                await connection.close()
            except EXPECTED_EXCEPTIONS as err:
                logger.exception(
                    "Failure closing RabbitMQ connection: %s. Moving on....", str(err)
                )

    async def close(self):
        """Close our RabbitMQ connection and all tasks.

        This is not part of the Channel Layer Specification; but it should be a
        part of every server's lifecycle.

        After calling this, no more messages will be delivered. All pending and
        future calls to `receive()` will raise `StopConsumer` (causing all
        WebSockets clients to disconnect). Undelivered messages will be lost.

        After this method returns, the RabbitMQ connection will be closed and
        all internal tasks will be completed.
        """

        if not self._want_close:
            self._want_close = True
            self._multi_queue.close()

            if hasattr(self, "_expire_task"):
                self._expire_task.cancel()
            if (
                hasattr(self, "_reconnect_forever_task")
                and self._carehare_connection.done()
            ):
                await self._carehare_connection.result().close()

        if hasattr(self, "_expire_task"):
            try:
                await self._expire_task
            except asyncio.CancelledError:
                pass
            if hasattr(self, "_expire_task"):
                delattr(self, "_expire_task")

        if hasattr(self, "_reconnect_forever_task"):
            await self._reconnect_forever_task
            if hasattr(self, "_reconnect_forever_task"):
                delattr(self, "_reconnect_forever_task")

    async def send(self, channel, message):
        """Send a message onto a (general or specific) channel."""
        assert self.valid_channel_name(channel), "Channel name not valid"
        _assert_channel_is_reply_channel(channel)

        logger.debug("send %r to %s", message, channel)

        body = serialize_message(ChannelRecipient(channel), message)

        connection = await self.carehare_connection

        try:
            await connection.publish(
                body, exchange_name="", routing_key=channel.split("!")[0]
            )
        except carehare.ServerSentNack:
            raise ChannelFull from None
        except carehare.ConnectionClosed:
            logger.warning(
                "Failed to send message %r to %s because we disconnected",
                message,
                channel,
            )

    async def receive(self, channel):
        """Receive the first message that arrives on the channel.

        If more than one coroutine waits on the same channel, only one waiter
        will receive the message when it arrives.
        """
        # Make sure the channel name is valid then get the non-local part
        # and thus its index
        assert self.valid_channel_name(channel)
        _assert_channel_is_reply_channel(channel)
        assert channel.startswith(self._queue_name + "!")
        return await self._multi_queue.get(channel)

    async def new_channel(self, prefix=""):
        """Create a new channel name that can be used by something in our
        process as a specific channel.
        """
        return "%s!%s" % (self._queue_name, _random_letters(12))

    async def group_add(self, group, channel):
        """Add the channel name to a group.

        Spec deviation: `channel` must have been created on the same event loop
        as the RabbitMQ connection. In other words: you can't subscribe someone
        else's channel to a group.
        """
        assert self.valid_group_name(group), "Group name not valid"
        assert self.valid_channel_name(channel), "Channel name not valid"

        # Connect, to make sure the next line doesn't happen during _setup_subscriptions()
        connection = await self.carehare_connection

        # Now, assume _setup_subscriptions() isn't running, so there's no race
        # writing+reading self._multi_queue.
        n_bindings = self._multi_queue.group_add(group, channel)
        if n_bindings == 1:
            logger.debug("Binding queue %s to group %s", self._queue_name, group)
            # This group is new to our connection-level queue. Make a
            # connection-level binding.

            try:
                await connection.queue_bind(
                    queue_name=self._queue_name,
                    exchange_name=self.groups_exchange,
                    routing_key=group,
                )
            except carehare.ChannelClosedByServer as err:
                logger.exception(
                    "RabbitMQ refused to bind: %s. Messages to group %s will not reach clients on this server.",
                    str(err),
                    group,
                )
            except carehare.ConnectionClosed:
                pass  # we'll queue_bind() during reconnect

    async def group_discard(self, group, channel):
        """Remove the channel from the named group if it is in the group;
        does nothing otherwise (does not error)

        Spec deviation: `channel` must have been created on the same event loop
        as the RabbitMQ connection. In other words: you can't subscribe someone
        else's channel to a group.
        """
        assert self.valid_group_name(group), "Group name not valid"
        assert self.valid_channel_name(channel), "Channel name not valid"

        # Connect, to make sure the next line doesn't happen during _setup_subscriptions()
        connection = await self.carehare_connection

        # Now, assume _setup_subscriptions() isn't running, so there's no race
        # writing+reading self._multi_queue.
        n_bindings = self._multi_queue.group_discard(group, channel)
        if n_bindings == 0:
            logger.debug("Unbinding queue %s from group %s", self._queue_name, group)
            try:
                await connection.queue_unbind(
                    queue_name=self._queue_name,
                    exchange_name=self.groups_exchange,
                    routing_key=group,
                )
            except carehare.ChannelClosedByServer as err:
                logger.exception(
                    "RabbitMQ refused to unbind from group %s: %s. This suggests a bug somewhere.",
                    str(err),
                    group,
                )
            except carehare.ConnectionClosed:
                pass  # after reconnect, we're unbound anyway

    async def group_send(self, group, message):
        """Send a message to the entire group."""
        assert self.valid_group_name(group), "Group name not valid"

        logger.debug("group_send %r to %s", message, group)

        body = serialize_message(GroupRecipient(group), message)

        connection = await self.carehare_connection

        try:
            await connection.publish(
                body, exchange_name=self.groups_exchange, routing_key=group
            )
        except carehare.ServerSentNack:
            # "Sending to a group never raises ChannelFull; instead, it must
            # silently drop the message if it is over capacity, as per ASGI’s
            # at-most-once delivery policy."
            # https://channels.readthedocs.io/en/stable/channel_layer_spec.html#capacity
            #
            # ... let's at least _warn_....
            logger.warning("Aborting send to group %s: a queue is at capacity", group)
        except carehare.ConnectionClosed:
            logger.warning(
                "Failed to send message %r to group %s because we disconnected",
                message,
                group,
            )



from __future__ import annotations

import asyncio
import logging
from typing import Any, Callable, Dict

from channels.exceptions import StopConsumer

from .util import ChannelRecipient, GroupRecipient, Recipient

BackpressureWarningInterval = 5.0  # seconds
ExpiryWarningInterval = 5.0  # seconds

logger = logging.getLogger(__name__)


class _Message:
    """Message passing through the MultiQueue."""

    data: Dict[str, Any]
    """The message to deliver."""

    n_queues: int
    """Number of queues where this message is waiting.

    If the message is waiting in 0 queues, it must be acked.
    """

    time: float
    """Absolute time when this message was queued.

    This is event-loop time, as used in `loop.call_at()`. See
    https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.call_at
    """

    ack: Callable[[], None]
    """Function to call exactly once, after the message has reached a consumer.

    The function must be a no-op after disconnect. (It mustn't send `ack` over
    another connection.) It may silently drop the ack -- but only when the
    connection is closing or closed. (Once the connection closes, the queue will
    disappear; that will implicitly ack this message.)
    """

    delivery_tag: int
    """Argument to ack()."""

    def __init__(
        self,
        data: Dict[str, Any],
        n_queues: int,
        time: float,
        ack: Callable[[], None],
    ):
        self.data = data
        self.n_queues = n_queues
        self.time = time
        self.ack = ack

    def mark_delivered(self):
        self.n_queues -= 1
        if self.n_queues == 0:
            self.ack()


class _ChannelQueue:
    """An "inner" queue: a single Django-Channels channel's _Messages.

    This solves a problem in the Django Channels Specification: a channel
    can exist without the layer knowing about it.

    Steps to reproduce the problem:

        1. Sender: send a message
        2. Recipient: receive the message

    This is backwards: the sender sent a message before the recipient
    even existed! But yep, that's the spec. The recipient _polls_ for
    messages, so it doesn't exist when it's busy or during startup.

    Hence _ChannelQueue: it's like asyncio.queue(), but with a notion of
    "unused". An unused queue is an empty queue with no getter.
    """

    def __init__(self):
        self._queue: asyncio.Queue[_Message] = asyncio.Queue()

    def unused(self) -> bool:
        """True if there are no queued messages and no pending getters."""
        return self._queue.empty() and not self._queue._getters

    def put_nowait(self, message: _Message) -> None:
        self._queue.put_nowait(message)

    async def get(self) -> Dict[str, Any]:
        """Wait for a message to become available, then return it."""
        message = await self._queue.get()
        message.mark_delivered()
        return message.data

    def remove_older_than(self, message_time: float) -> bool:
        """Remove (and mark_delivered()) _Messages we won't use.

        Return True if anything was deleted.
        """
        retval = False
        # We need to peek: self._queue is an asyncio.Queue, and
        # self._queue._queue is the collections.deque() _inside_ it.
        while not self._queue.empty() and self._queue._queue[0].time < message_time:
            message = self._queue._queue.popleft()
            message.mark_delivered()
            retval = True

        return retval

    @property
    def oldest_message_time(self) -> float:
        """The earliest "expires" of all queued messages."""
        return self._queue._queue[0].time

    def close(self) -> None:
        """Raise StopConsumer on all readers."""
        for waiter in self._queue._getters:
            if not waiter.done():
                waiter.set_exception(StopConsumer("Django Channel is shutting down"))


class MultiQueue:
    """Asyncio-friendly one-to-many queue that blocks when capacity is reached.

    Usage:

        await multi_queue.put_nowait("q1", Message({"foo": "hey"}))
        await multi_queue.put_nowait("q2", Message({"foo": "jude"}))
        await multi_queue.get("q2")  # => {"foo": "jude"}

    Ordering guarantee: on a given queue, messages are retrieved in FIFO order.

    MultiQueue gives back-pressure. Messages are only acked when delivered.
    RabbitMQ will never send us more than `prefetch_count` un-acked messages,
    so the MultiQueue will never hold more than `prefetch_count` messages.
    Excess messages will queue within RabbitMQ itself; depending on
    `remote_capacity`, this backlog may lead senders to raise `ChannelFull`.
    When MultiQueue holds `prefetch_count` messages, we log a warning.

    MultiQueue also manages local groups. "Groups" are a distributed concept:
    10 channels may all be subscribed to a group, across three servers. Here,
    we deal with the 3-4 channels on this server which are subscribed to the
    group.

    To prevent overflow, each message has an expiry time. We expire old messages
    to make room for new ones. In this event, we ack the old messages and log
    a warning.
    """

    def __init__(self, capacity):
        self.capacity = capacity
        self.n = 0

        self._local_groups = {}  # group => {channel, ...}
        self._channels = {}  # channel => _ChannelQueue
        self._nonempty = asyncio.Event()
        self._closed = asyncio.Event()
        # We'll order puts using a lock. Waiters on an asyncio.Lock are
        # released in FIFO order.
        self._putter_lock = asyncio.Lock()
        self._last_logged_backpressure = 0  # loop.time() result
        self._last_logged_expiry = 0  # loop.time() result

    def _drop_all_older_than(self, message_time: float) -> bool:
        """Drop expired messages; return `true` if something happened."""
        deleted = False
        channels_to_delete = []
        for channel, queue in self._channels.items():
            if queue.remove_older_than(message_time):
                deleted = True
            if queue.unused():
                channels_to_delete.append(channel)
        for channel in channels_to_delete:
            del self._channels[channel]
        return deleted

    async def expire_locally_until_closed(self, local_expiry: float):
        loop = asyncio.get_running_loop()
        closed_task = asyncio.create_task(self._closed.wait())

        while not closed_task.done():
            while not self._nonempty.is_set():
                nonempty_task = asyncio.create_task(self._nonempty.wait())
                await asyncio.wait(
                    {nonempty_task, closed_task},
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if closed_task.done():
                    return

            # Assume messages expire in FIFO order. Therefore, we know there's
            # no need to check for expiry until the first unhandled message
            # expires.
            oldest_message_time = min(
                [
                    queue.oldest_message_time
                    for queue in self._channels.values()
                    if not queue._queue.empty()
                ]
            )
            now = loop.time()
            if oldest_message_time > now - local_expiry:
                await asyncio.wait(
                    {closed_task}, timeout=oldest_message_time + local_expiry - now
                )
                if closed_task.done():
                    return
                now = loop.time()

            dropped = self._drop_all_older_than(now - local_expiry)
            if dropped:
                self._log_local_expiry_debounced(now)

    def _increase_n(self):
        assert self.n < self.capacity
        self.n += 1  # decreased in self._cleanup_message()
        if self.n == self.capacity:
            now = asyncio.get_running_loop().time()
            self._log_backpressure_debounced(now)
        self._nonempty.set()

    def _decrease_n(self) -> None:
        self.n -= 1
        if self.n == 0:
            self._nonempty.clear()

    def put_nowait(
        self,
        recipient: Recipient,
        data: Dict[str, Any],
        time: float,
        ack: Callable[[], None],
    ) -> None:
        """Queue a message.

        This will call `ack` when the message is either delivered (to all
        recipients) or expired.

        For a GroupRecipient, recipient channels are chosen at call time. When
        a message is delivered, the group's members may have changed. A
        consumer calling `get()` may receive a message after it called
        `group_discard()` -- as long as the `put_nowait()` happened before
        `group_discard()`.
        """
        if isinstance(recipient, ChannelRecipient):
            channels = [recipient.channel]
        else:
            assert isinstance(recipient, GroupRecipient)
            try:
                channels = self._local_groups[recipient.group]
            except KeyError:
                ack()
                return

        def ack_and_decrease_n():
            ack()
            self._decrease_n()

        message = _Message(data, len(channels), time, ack=ack_and_decrease_n)

        for channel in channels:
            if channel not in self._channels:
                self._channels[channel] = _ChannelQueue()
            self._channels[channel].put_nowait(message)

        self._increase_n()

    def _build_big_queues_str(self) -> str:
        blockers = [(name, len(q._queue._queue)) for name, q in self._channels.items()]
        blockers.sort(key=lambda b: b[1], reverse=True)
        return ", ".join(f"{name} ({count})" for name, count in blockers[:3])

    def _log_backpressure_debounced(self, now: float) -> None:
        """Log back-pressure if we haven't logged it in a while.

        Back-pressure is by design: the user configured it.

        However, back-pressure on a well-oiled website often means something's
        wrong -- for example, a cancelled Websocket client erroneously still
        added to a group; or a user with a slow web browser only consuming one
        message every 5s. You can't rely on a ChannelFull exception appearing:
        back-pressure can convince your users the site is broken, so they'll go
        away.

        Should this be _info_ (because it's a normal event) or _warning_?
        Warning seems appropriate because back-pressure can mean disaster on a
        production website.
        """
        if now - self._last_logged_backpressure > BackpressureWarningInterval:
            logger.warning(
                "Back-pressuring. Biggest queues: %s", self._build_big_queues_str()
            )
            self._last_logged_backpressure = now

    def _log_local_expiry_debounced(self, now: float) -> None:
        """Log dropped messages.

        Expiry is by design: the user configured it. But it's usually not
        intended.

        Should this be _info_ (because it's a normal event) or _warning_?
        Warning seems appropriate because expiry usually means there's a
        problem the dev should fix.
        """
        if now - self._last_logged_expiry > ExpiryWarningInterval:
            logger.warning(
                "A message (or messages) expired locally. Biggest queues: %s",
                self._build_big_queues_str(),
            )
            self._last_logged_expiry = now

    async def get(self, channel: str) -> Any:
        """Maybe wait for a `put_channel()` on `channel`; return its data."""
        if self._closed.is_set():
            raise StopConsumer("Django Channel was shut down")

        if channel not in self._channels:
            self._channels[channel] = _ChannelQueue()

        try:
            return await self._channels[channel].get()
        finally:  # Even if it raises StopConsumer
            if (
                channel in self._channels  # it may have been deleted in await?
                and self._channels[channel].unused()
            ):
                del self._channels[channel]

    def group_add(self, group, channel):
        channels = self._local_groups.setdefault(group, set())
        channels.add(channel)
        return len(channels)

    def group_discard(self, group, channel):
        """Remove `channel` from `group` and return n_channels_remaining.

        Return None if the channel is not in the group.
        """
        if group not in self._local_groups:
            return None  # don't create set

        channels = self._local_groups[group]
        try:
            channels.remove(channel)
        except KeyError:
            return None  # it was already removed

        ret = len(channels)

        if ret == 0:
            del self._local_groups[group]

        return ret

    def close(self):
        """Nullify pending puts; raise StopConsumer on pending gets."""
        if self._closed.is_set():
            return
        self._closed.set()

        # Cancel all gets
        for out_queue in self._channels.values():
            out_queue.close()
        self._channels.clear()



import asyncio
from typing import Any, Dict, Iterable, NamedTuple, Union

import msgpack


class ChannelRecipient(NamedTuple):
    channel: str


class GroupRecipient(NamedTuple):
    group: str


Recipient = Union[ChannelRecipient, GroupRecipient]


def serialize_message(recipient: Recipient, data: Dict[str, Any]) -> bytes:
    assert isinstance(data, dict), "data is not a dict"
    assert "__asgi_channel__" not in data
    assert "__asgi_group__" not in data
    augmented_data = dict(data)
    if isinstance(recipient, ChannelRecipient):
        augmented_data["__asgi_channel__"] = recipient.channel
    elif isinstance(recipient, GroupRecipient):
        augmented_data["__asgi_group__"] = recipient.group
    return msgpack.packb(augmented_data)


class DeserializeResult(NamedTuple):
    recipient: Recipient
    data: Dict[str, Any]


def deserialize_message(message_bytes: bytes) -> DeserializeResult:
    data = msgpack.unpackb(message_bytes)
    channel = data.pop("__asgi_channel__", None)
    group = data.pop("__asgi_group__", None)
    assert (channel is None) != (group is None)
    if channel is not None:
        recipient = ChannelRecipient(channel)
    else:
        recipient = GroupRecipient(group)
    return DeserializeResult(recipient, data)


async def gather_without_leaking(tasks: Iterable[asyncio.Task]):
    """Run a bunch of tasks to completion, _then_ raise the first exception.

    This differs from regular `asyncio.gather()`, which leaves tasks running on
    the event loop without waiting for them to finish.

    It also differs because it accepts a list, not varargs.

    This looks like a hack: shouldn't it be part of asyncio instead? Why isn't
    this the _default_? ... dunno.
    """
    try:
        # Run, raising first exception
        await asyncio.gather(*tasks)
    except Exception:
        # Wait for all tasks to finish, exceptional or not
        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        raise



import asyncio
from functools import partial

import carehare

from .multiqueue import MultiQueue
from .util import deserialize_message


async def consume_into_multi_queue_until_connection_close(
    connection: carehare.Connection,
    channel: str,
    multi_queue: MultiQueue,
    prefetch_count: int,
) -> None:
    loop = asyncio.get_running_loop()

    try:
        async with connection.acking_consumer(
            channel, prefetch_count=prefetch_count
        ) as consumer:

            def _ack_or_no_op(delivery_tag: int):
                try:
                    consumer.ack(delivery_tag)
                except OSError:  # asyncio.Transport.write() failed
                    pass

            while True:
                # back-pressure until `consumer.ack()` was called enough times
                try:
                    body, delivery_tag = await consumer.next_delivery()
                except carehare.ConnectionClosed:
                    break
                recipient, data = deserialize_message(body)

                multi_queue.put_nowait(
                    recipient,
                    data,
                    time=loop.time(),
                    ack=partial(_ack_or_no_op, delivery_tag),
                )
    except carehare.ConnectionClosed:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['channels_rabbitmq']

package_data = \
{'': ['*']}

install_requires = \
['carehare>=1.0,<2.0', 'channels>=3.0,<5', 'msgpack>=1.0,<2.0']

setup_kwargs = {
    'name': 'channels-rabbitmq',
    'version': '4.0.1',
    'description': 'RabbitMQ-based ASGI channel layer implementation',
    'long_description': 'channels_rabbitmq\n=================\n\nA Django Channels channel layer that uses RabbitMQ as its backing store.\n\nDoes not support `Worker and Background Tasks\n<https://channels.readthedocs.io/en/stable/topics/worker.html>`_.\n(See `Rationale\n<https://github.com/CJWorkbench/channels_rabbitmq/pull/11#issuecomment-499185070>`_\nand use ``await get_channel_layer().current_connection`` to send to job queues.)\n\nWorks with Python 3.8 or 3.9.\n\nInstallation\n------------\n\n``pip install channels_rabbitmq``\n\nUsage\n-----\n\nThen set up the channel layer in your Django settings file like so::\n\n    CHANNEL_LAYERS = {\n        "default": {\n            "BACKEND": "channels_rabbitmq.core.RabbitmqChannelLayer",\n            "CONFIG": {\n                "host": "amqp://guest:guest@127.0.0.1/asgi",\n                # "ssl_context": ... (optional)\n            },\n        },\n    }\n\nPossible options for ``CONFIG`` are listed below.\n\n``host``\n~~~~~~~~\n\nURL of the server to connect to, adhering to `RabbitMQ spec\n<https://www.rabbitmq.com/uri-spec.html>`_. To connect to a RabbitMQ cluster,\nuse a DNS server to resolve a hostname to multiple IP addresses.\nchannels_rabbitmq will automatically reconnect if at least one of them is\nreachable in case of a disconnection.\n\n``expiry``\n~~~~~~~~~~\n\nMinimum number of seconds a message should wait in a RabbitMQ queue, before it\nmay be silently dropped.\n\nDefaults to ``60``. You generally shouldn\'t need to change this, but you may\nwant to turn it down if you have peaky traffic you wish to drop, or up if you\nhave peaky traffic you want to backlog until you get to it.\n\n``local_capacity``\n~~~~~~~~~~~~~~~~~~\n\nNumber of incoming messages queued in memory. Defaults to ``100``. (A message\nsent to a group with two channels counts as one message.) When ``local_capacity``\nmessages are queued, the message backlog will grow on RabbitMQ.\n\n(This controls the ``prefetch_count`` on the RabbitMQ queue.)\n\n``local_expiry``\n~~~~~~~~~~~~~~~~\n\nMinimum number of seconds a message received from RabbitMQ must be held in\nmemory waiting for ``receive()``, before it may be dropped. Defaults to\n``expiry``.\n\nA warning will be logged when a message expires locally. The warning can\nindicate that a channel has more messages than it can handle; or that\nmessages are being sent to a channel that does not exist. (Perhaps a missing\nchannel was implied by ``group_add()``, and a matching ``group_discard()``\nwas never called.)\n\nIf ``local_expiry < expiry``, then you can end up ignoring (and logging)\nmessages locally while they still exist in the RabbitMQ queue. These messages\nwill be acked, so RabbitMQ will behave as though they were delivered.\n\n``remote_capacity``\n~~~~~~~~~~~~~~~~~~~\n\nNumber of messages stored on RabbitMQ for each client. Defaults to ``100``.\n(A message sent to a group with three channels on two distinct clients counts\nas two messages.) When ``remote_capacity`` messages are queued in RabbitMQ,\nthe channel will refuse new messages. Calls from any client to ``send()`` or\n``group_send()`` to the at-capacity client will raise ``ChannelFull``.\n\n``ssl_context``\n~~~~~~~~~~~~~~~\n\nAn `SSL context\n<https://docs.python.org/3/library/ssl.html#ssl-contexts>`_. Changes the\ndefault ``host`` port to 5671 (instead of 5672).\n\nFor instance, to connect to an TLS RabbitMQ service that will verify your\nclient::\n\n    import ssl\n    ssl_context = ssl.create_default_context(\n        cafile=str(Path(__file__).parent.parent / \'ssl\' / \'server.cert\'),\n    )\n    ssl_context.load_cert_chain(\n        certfile=str(Path(__file__).parent.parent / \'ssl\' / \'client.certchain\'),\n        keyfile=str(Path(__file__).parent.parent / \'ssl\' / \'client.key\'),\n    )\n    CHANNEL_LAYERS[\'default\'][\'CONFIG\'][\'ssl_context\'] = ssl_context\n\nBy default, there is no SSL context; all messages (and passwords) are\nare transmitted in cleartext.\n\n``groups_exchange``\n~~~~~~~~~~~~~~~~~~~\n\nGlobal direct exchange name used by channels to exchange group messages.\nDefaults to ``"groups"``. See also `Design decisions`_.\n\nAccessing Carehare\n------------------\n\nWe use `carehare`_ for its thorough handling of errors.\n\nDjango Channels\' specification does not account for "connecting" and\n"disconnecting". This layer does its best by constantly reconnecting, forever.\n\nCall ``await get_channel_layer().current_connection`` to access an open Carehare\nconnection. This lets you use job queues without "Worker and Background Tasks".\nLike this::\n\n    # raise asyncio.CancelledError on failure\n    connection = await get_channel_layer().carehare_connection\n\n    # raise carehare.ConnectionClosed or carehare.ChannelClosed on error\n    await connection.publish(b"task", routing_key="job_queue")\n\n(The Carehare documentation explains how to build workers.)\n\nA note on errors: a "connected" connection isn\'t guaranteed to *stay* connected\nthroughout every publish. It was merely connected *at some point in the past*.\nWhen a disconnect occurs, all pending operations on that connection will raise\n``carehare.ConnectionClosed``. This channel layer will log the error, and\n``get_channel_layer().carehare_connection`` will point to a new Future. (This\nerror+reconnect is *guaranteed to happen* in production.)\n\nPublish messages from Celery\n----------------------------\n\nMany `Celery\n<https://docs.celeryproject.org/en/stable/getting-started/introduction.html>`_\nusers want to send messages to websockets users.\n\nThis is doable, though not intuitive. Don\'t use any Django Channels code:\nChannels layers depend on long-running connections, and Celery bans those.\nSame goes for carehare: don\'t use it from Celery.\n\nInstead, from a Celery-worker ``@task`` you can send messages to your\nDjango-Channels consumers using Celery\'s RabbitMQ connection::\n\n    from typing import Any, Dict\n\n    import msgpack\n\n    def publish_message_to_group(message: Dict[str, Any], group: str) -> None:\n        with current_app.producer_pool.acquire(block=True) as producer:\n            producer.publish(\n                msgpack.packb({\n                  "__asgi_group__": group,\n                  **message,\n                }),\n                exchange="groups",  # groups_exchange\n                content_encoding="binary",\n                routing_key=group,\n                retry=False,  # Channel Layer at-most once semantics\n            )\n\nTo call it, from a Celery-worker ``@task``...::\n\n    publish_message_to_group({ "type": "chat.message", "text": "hi" }, "a-group")\n\n... and a Django-Channels consumer like this will receive it::\n\n    class WebsocketConnectionConsumer(AsyncWebsocketConsumer):\n        async def connect(self):\n            await self.channel_layer.group_add("a-group", self.channel_name)\n\n        async def disconnect(self):\n            await self.channel_layer.group_discard("a-group", self.channel_name)\n\n        async def chat_message(self, event):\n            assert event["text"] == "hi"\n\nAlternatively, write your workers asynchronously, directly in `carehare`_.\nIt\'s more lightweight and faster than Celery, and the error handling is\nsimpler.\n\nDesign decisions\n----------------\n\nTo scale enormously, this layer only creates one RabbitMQ queue per instance.\nThat means one web server gets one RabbitMQ queue, no matter how many\nwebsocket connections are open. For each message being sent, the client-side\nlayer determines the RabbitMQ queue name and uses it as the routing key.\n\nGroups are implemented using a single, global RabbitMQ direct exchange called\n"groups" by default. To send a message to a group, the layer sends the message\nto the "groups" exchange with the group name as the routing key. The client\nbinds and unbinds during ``group_add()`` and ``group_remove()`` to ensure\nmessages for any of its groups will reach it. See also the `groups_exchange`_\noption.\n\nRabbitMQ queues are ``exclusive``: when a client disconnects (through close or\ncrash), RabbitMQ will delete the queue and unbind the groups.\n\nOnce a connection has been created, it pollutes the event loop so that\n``async_to_sync()`` will destroy the connection if it was created within\n``async_to_sync()``. Each connection starts a background async loop that pulls\nmessages from RabbitMQ and routes them to receiver queues; each ``receive()``\nqueries receiver queues. Empty queues with no connections are deleted.\n\nDeviations from the Channel Layer Specification\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nThe `Channel Layer Specification\n<https://channels.readthedocs.io/en/latest/channel_layer_spec.html>`_ bends to\nRedis-related restrictions. RabbitMQ cannot emulate Redis. Here are the\ndifferences:\n\n* **No ``flush`` extension**: To flush all state, simply disconnect all clients.\n  (RabbitMQ won\'t allow one client to delete another client\'s data structures.)\n* **No ``group_expiry`` option**: The `group_expiry option\n  <https://channels.readthedocs.io/en/latest/channel_layer_spec.html#persistence>`_\n  recovers when a ``group_add()`` has no matching ``group_discard()``. But the\n  "group membership expiry" logic has a fatal flaw: it disconnects legitimate\n  members. ``channels_rabbitmq`` addresses each root problem instead:\n\n  * Web-server crash: RabbitMQ wipes all state related to a web server when\n    the web server disconnects. There\'s no problem here for ``group_expiry``\n    to solve.\n  * Programming errors: You may err and call ``group_add()`` without\n    eventually calling ``group_discard()``. Redis can\'t detect this\n    programming error (because it can\'t detect web-server crashes). RabbitMQ\n    can. The ``local_expiry`` option keeps your site running after you\n    erroneously miss a ``group_discard()``. The channel layer warns when\n    discarding expired messages. Monitor your server logs to detect your\n    errors.\n* **No "normal channels"**: `Normal channels\n  <https://channels.readthedocs.io/en/latest/channel_layer_spec.html#channels>`_\n  are job queues. In most projects, "normal channel" readers are worker\n  processes, ideally divorced from Websockets and Django.\n\n  If you want an async, RabbitMQ-based job queue, investigate `carehare\n  <https://github.com/CJWorkbench/carehare>`_.\n\n  If you\'re using Celery with the same RabbitMQ server, you can `publish\n  messages from Celery`_, too.\n\nDependencies\n------------\n\nYou\'ll need Python 3.8+ and a RabbitMQ server.\n\nIf you have Docker, here\'s how to start a development server::\n\n    ssl/prepare-certs.sh  # Create SSL certificates used in tests\n    docker run --rm -it \\\n         -p 5671:5671 \\\n         -p 5672:5672 \\\n         -p 15672:15672 \\\n         -v "/$(pwd)"/ssl:/ssl \\\n         -e RABBITMQ_SSL_CACERTFILE=/ssl/ca.cert \\\n         -e RABBITMQ_SSL_CERTFILE=/ssl/server.cert \\\n         -e RABBITMQ_SSL_KEYFILE=/ssl/server.key \\\n         -e RABBITMQ_SSL_VERIFY=verify_peer \\\n         -e RABBITMQ_SSL_FAIL_IF_NO_PEER_CERT=true \\\n         rabbitmq:3.7.8-management-alpine\n\nYou can access the RabbitMQ management interface at http://localhost:15672.\n\nContributing\n------------\n\nTo add features and fix bugs\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nFirst, start a development RabbitMQ server::\n\n    ssl/prepare-certs.sh  # Create SSL certificates used in tests\n    docker run --rm -it \\\n         -p 5671:5671 \\\n         -p 5672:5672 \\\n         -p 15672:15672 \\\n         -v "/$(pwd)"/ssl:/ssl \\\n         -e RABBITMQ_SSL_CACERTFILE=/ssl/ca.cert \\\n         -e RABBITMQ_SSL_CERTFILE=/ssl/server.cert \\\n         -e RABBITMQ_SSL_KEYFILE=/ssl/server.key \\\n         -e RABBITMQ_SSL_VERIFY=verify_peer \\\n         -e RABBITMQ_SSL_FAIL_IF_NO_PEER_CERT=true \\\n         rabbitmq:3.8.11-management-alpine\n\nNow take on the development cycle:\n\n#. ``tox`` # to ensure tests pass.\n#. Write new tests in ``tests/`` and make sure they fail.\n#. Write new code in ``channels_rabbitmq/`` to make the tests pass.\n#. Submit a pull request.\n\nTo deploy\n~~~~~~~~~\n\nUse `semver <https://semver.org/>`_.\n\n#. ``git push`` and make sure Travis tests all pass.\n#. ``git tag vX.X.X``\n#. ``git push --tags``\n\nTravisCI will push to PyPi.\n',
    'author': 'Adam Hooper',
    'author_email': 'adam@adamhooper.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
